/*
 *  -- MAGMA (version 1.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     November 2011
 *
 *  @generated c Tue Mar 12 12:57:57 2013
 *
 **/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

#define PRECISION_c

int main(int argc, char **argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }

    real_Double_t   gflops, magma_perf, magma_time, mic_perf, mic_time, cpu_perf, cpu_time;
    float          magma_error, mic_error, work[1];
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t M, N, Xm, Ym, lda, sizeA, sizeX, sizeY;
    magma_int_t incx = 1;
    magma_int_t incy = 1;
    magmaFloatComplex c_neg_one = MAGMA_C_NEG_ONE;
    magmaFloatComplex alpha = MAGMA_C_MAKE(  1.5, -2.3 );
    magmaFloatComplex beta  = MAGMA_C_MAKE( -0.6,  0.8 );
    magmaFloatComplex *A, *X, *Y, *Ymic, *Ymagma;
    magmaFloatComplex *dA, *dX, *dY;
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );

    printf("    M     N   MAGMA Gflop/s (ms)  MIC Gflop/s (ms)   CPU Gflop/s (ms)  MAGMA error  MIC error\n");
    printf("===================================================================================================\n");
    for( int i = 0; i < opts.ntest; ++i ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            M = opts.msize[i];
            N = opts.nsize[i];
            lda    = ((M+31)/32)*32;
            gflops = FLOPS_CGEMV( M, N ) / 1e9;

            if ( opts.transA == MagmaNoTrans ) {
                Xm = N;
                Ym = M;
            } else {
                Xm = M;
                Ym = N;
            }

            sizeA = lda*N;
            sizeX = incx*Xm;
            sizeY = incy*Ym;
            
            TESTING_MALLOC_HOST( A, magmaFloatComplex, sizeA );
            TESTING_MALLOC_HOST( X, magmaFloatComplex, sizeX );
            TESTING_MALLOC_HOST( Y, magmaFloatComplex, sizeY );
            TESTING_MALLOC_HOST( Ymic, magmaFloatComplex, sizeY );
            TESTING_MALLOC_HOST( Ymagma,  magmaFloatComplex, sizeY );
            
            TESTING_MALLOC_DEV( dA, magmaFloatComplex, sizeA );
            TESTING_MALLOC_DEV( dX, magmaFloatComplex, sizeX );
            TESTING_MALLOC_DEV( dY, magmaFloatComplex, sizeY );
            
            /* Initialize the matrix */
            lapackf77_clarnv( &ione, ISEED, &sizeA, A );
            lapackf77_clarnv( &ione, ISEED, &sizeX, X );
            lapackf77_clarnv( &ione, ISEED, &sizeY, Y );
            
            /* =====================================================================
               Performs operation using MIC
               =================================================================== */
            magma_csetmatrix( M, N, A, 0, lda, dA, 0, lda, queue );
            //magma_csetvector( Xm, X, incx, dX, incx );
            //magma_csetvector( Ym, Y, incy, dY, incy );
            magma_csetmatrix( Xm, 1, X, 0, Xm, dX, 0, Xm, queue );
            magma_csetmatrix( Ym, 1, Y, 0, Ym, dY, 0, Ym, queue );
            
            mic_time = magma_sync_wtime( 0 );
            magma_cgemv( opts.transA, M, N, alpha, dA, 0, lda, dX, 0, incx, beta, dY, 0, incy, queue );
            mic_time = magma_sync_wtime( 0 ) - mic_time;
            mic_perf = gflops / mic_time;
            
            //magma_cgetvector( Ym, dY, incy, Ymic, incy );
            magma_cgetmatrix( Ym, 1, dY, 0, Ym, Ymic, 0, Ym, queue );
            
            /* =====================================================================
               Performs operation using MAGMABLAS
               =================================================================== */
            //magma_csetvector( Ym, Y, incy, dY, incy );
            magma_csetmatrix( Ym, 1, Y, 0, Ym, dY, 0, Ym, queue );
            
            magma_time = magma_sync_wtime( 0 );
            magma_cgemv( opts.transA, M, N, alpha, dA, 0, lda, dX, 0, incx, beta, dY, 0, incy, queue );
            magma_time = magma_sync_wtime( 0 ) - magma_time;
            magma_perf = gflops / magma_time;
            
            //magma_cgetvector( Ym, dY, incx, Ymagma, incx );
            magma_cgetmatrix( Ym, 1, dY, 0, Ym, Ymagma, 0, Ym, queue );
            
            /* =====================================================================
               Performs operation using CPU BLAS
               =================================================================== */
            cpu_time = magma_wtime();
            blasf77_cgemv( lapack_const(opts.transA), &M, &N,
                           &alpha, A, &lda,
                                   X, &incx,
                           &beta,  Y, &incy );
            cpu_time = magma_wtime() - cpu_time;
            cpu_perf = gflops / cpu_time;
            
            /* =====================================================================
               Check the result
               =================================================================== */
            blasf77_caxpy( &Ym, &c_neg_one, Y, &incy, Ymagma, &incy );
            magma_error = lapackf77_clange( "M", &Ym, &ione, Ymagma, &Ym, work ) / Ym;
            
            blasf77_caxpy( &Ym, &c_neg_one, Y, &incy, Ymic, &incy );
            mic_error = lapackf77_clange( "M", &Ym, &ione, Ymic, &Ym, work ) / Ym;
            
            printf("%5d %5d   %7.2f (%7.2f)   %7.2f (%7.2f)	%7.2f (%7.2f)    %8.2e     %8.2e\n",
                   (int) M, (int) N,
                   mic_perf, 1000.*mic_time,
                   magma_perf, 1000.*magma_time,
                   cpu_perf,    1000.*cpu_time,
                   magma_error, mic_error );
            
            TESTING_FREE_HOST( A );
            TESTING_FREE_HOST( X );
            TESTING_FREE_HOST( Y );
            TESTING_FREE_HOST( Ymic );
            TESTING_FREE_HOST( Ymagma );
            
            TESTING_FREE_DEV( dA );
            TESTING_FREE_DEV( dX );
            TESTING_FREE_DEV( dY );
        }
        if ( opts.niter > 1 ) {
            printf( "\n" );
        }
    }

    magma_queue_destroy( queue );
    magma_finalize();
    
    return 0;
}
