/*
    -- micMAGMA (version 0.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       April 2012

       @generated c Tue Mar 12 12:57:58 2013

*/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

// Flops formula
#define PRECISION_c
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(m, n) ( 6.*FMULS_GEQRF(m, n) + 2.*FADDS_GEQRF(m, n) )
#else
#define FLOPS(m, n) (    FMULS_GEQRF(m, n) +    FADDS_GEQRF(m, n) )
#endif

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cgeqrf
*/
int main( int argc, char** argv)
{
    
    real_Double_t    gflops, mic_perf, cpu_perf, mic_time, cpu_time;
    float           error, matnorm, work[1];
    magmaFloatComplex  mzone = MAGMA_C_NEG_ONE;
    magmaFloatComplex *h_A, *h_R, *tau, *hwork, tmp[1];
    magmaFloatComplex_ptr d_A;

    /* Matrix size */
    magma_int_t M = 0, N = 0, n2, lda, ldda, lhwork, alltest=21;
    magma_int_t size[21] = {960, 1920,3072,4032,4992,5952,7104,8064,9024,10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694};

    magma_int_t i, info, min_mn;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t checkres=0, nbtest=1, warmup=1;


    if (argc == 1){
        nbtest = alltest;
        printf("  testing_cgeqrf_mic running all tests sizes \n");
        M = N = size[alltest-1];
    }else{
        nbtest = 1;
    }


    for(i = 1; i<argc; i++){
        if (strcmp("-N", argv[i])==0){
            N = atoi(argv[++i]);
        }else if (strcmp("-M", argv[i])==0){
            M = atoi(argv[++i]);
        }else if ( strcmp("-c", argv[i]) == 0 ) {
            checkres = 1;
        }else if ( strcmp("-warmup", argv[i]) == 0 ) {
            warmup = atoi(argv[++i]);
        }else if ( strcmp("-nbtest", argv[i]) == 0 ) {
            nbtest = atoi(argv[++i]);
        }else if ( strcmp("-h", argv[i]) == 0 ) {
            printf("\nUsage: \n");
            printf("  testing_cgeqrf_mic -M %d -N %d\n\n", 1024, 1024);
            exit(1);
        }
    }
    if ( M <= 0 ) {
        if(N <=0 ){
            nbtest = alltest;
            printf("  testing_cgeqrf_mic running all tests sizes \n");
            M = N = size[alltest-1];
        }else{
            M = N;
        }
    }else if(N<=0){
            N = M;
    }

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }

    ldda  = ((M+31)/32)*32;
    ldda   = (ldda % 256 == 0) ? (ldda + 32) : ldda;
    n2     = ldda * N;
    min_mn = min(M, N);

    /* Allocate host memory for the matrix */
    TESTING_MALLOC(      tau, magmaFloatComplex, min_mn );
    TESTING_MALLOC(      h_A, magmaFloatComplex, n2     );
    TESTING_MALLOC_HOST( h_R, magmaFloatComplex, n2     );
    TESTING_MALLOC_DEV ( d_A, magmaFloatComplex, ldda*N );

    lhwork = -1;
    lapackf77_cgeqrf(&M, &N, h_A, &M, tau, tmp, &lhwork, &info);
    lhwork = (magma_int_t)MAGMA_C_REAL( tmp[0] );

    TESTING_MALLOC( hwork, magmaFloatComplex, lhwork );

    printf("\n\n");
    printf("  M     N    CPU GFlop/s (sec)   MIC GFlop/s (sec)   ||R||_F / ||A||_F\n");
    printf("======================================================================\n");

    
    for(i = 0; i < nbtest; i++){
        if(nbtest>1)    
            M = N = size[i];
        min_mn= min(M, N);
        lda   = M;
        n2    = lda * N;
        ldda  = ((M + 31) / 32) * 32;
                ldda   = (ldda % 256 == 0) ? (ldda + 32) : ldda;
        gflops = FLOPS( (float)M, (float)N ) * 1e-9;

        /* Initialize the matrix */
        lapackf77_clarnv( &ione, ISEED, &n2, h_A );
        lapackf77_clacpy( MagmaUpperLowerStr, &M, &N, h_A, &lda, h_R, &lda );

        /* =====================================================================
           Performs operation using LAPACK
           =================================================================== */
        if(checkres==1){
            cpu_time = get_time();
            lapackf77_cgeqrf(&M, &N, h_A, &M, tau, hwork, &lhwork, &info);
            cpu_time = get_time() - cpu_time;
            if (info < 0)
                printf("Argument %d of lapack_cgeqrf had an illegal value.\n", -info);
            cpu_perf = gflops / cpu_time;
        }

        /* ====================================================================
           Performs operation using MAGMA
           =================================================================== */
        for(magma_int_t w = 0; w < warmup; w++){
            magma_csetmatrix( M, N, h_R, 0, lda, d_A, 0, ldda, queue );
            mic_time = get_time();
            magma_cgeqrf2_mic( M, N, d_A, 0, ldda, tau, &info, queue);
            mic_time = get_time() - mic_time;
            if (info < 0)
                printf("Argument %d of magma_cgeqrf2 had an illegal value.\n", -info);
        }
        mic_perf = gflops / mic_time;

        /* =====================================================================
           Check the result compared to LAPACK
           =================================================================== */
        if(checkres==1){
                magma_cgetmatrix( M, N, d_A, 0, ldda, h_R, 0, M, queue );
                matnorm = lapackf77_clange("f", &M, &N, h_A, &M, work);
                blasf77_caxpy(&n2, &mzone, h_A, &ione, h_R, &ione);
                error = lapackf77_clange("f", &M, &N, h_R, &M, work) / matnorm;
        }
        printf("%5d %5d   %6.2f (%6.2f)     %6.2f (%6.2f)       %e\n",
                        M, N, cpu_perf, cpu_time, mic_perf, mic_time,
                        error);
        fflush(0);    
    }

        /* clean up */
        TESTING_FREE( tau );
        TESTING_FREE( h_A );
        TESTING_FREE( hwork );
        TESTING_FREE_HOST( h_R );
        TESTING_FREE_DEV( d_A );

        magma_queue_destroy( queue );
        magma_finalize();
}
