/*
 *  -- micMAGMA (version 0.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
 * @generated c Tue Mar 12 12:57:58 2013
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

#define PRECISION_c
// Flops formula
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(n) ( 6. * FMULS_POTRF(n) + 2. * FADDS_POTRF(n) )
#else
#define FLOPS(n) (      FMULS_POTRF(n) +      FADDS_POTRF(n) )
#endif

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cpotrf
*/
#define hA(i,j) hA[ i + j*lda ]

int main( int argc, char** argv) 
{
    real_Double_t gflops, mic_perf, cpu_perf, mic_time, cpu_time;
    magmaFloatComplex *hA, *hR;
    magmaFloatComplex_ptr dA;
    magma_int_t N = 0, flag, n2, lda, ldda;
    //magma_int_t size[24] = {960, 1920, 3072, 4032, 4992, 5952, 7104, 8064, 9024, 10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694, 27892, 27892, 27892};
    magma_int_t size[27] = {960, 1920, 3072, 2880, 3840, 4096, 4800, 5760, 6720, 7680, 8064, 8640, 9600, 10560, 12480, 14400, 14560, 16320, 18240, 20160, 22080, 24000, 25920, 26880, 27892, 27892, 27892};

    magma_int_t i, uplo, info;
    magmaFloatComplex mz_one = MAGMA_C_NEG_ONE;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    float      work[1], matnorm, diffnorm;
   
        uplo = MagmaLower;
        flag = 0;
    if (argc != 1){
        for(i = 1; i<argc; i++){        
            if (strcmp("-N", argv[i])==0) {
                N = atoi(argv[++i]);
                                flag = 1;
                        } else if (strcmp("-uplo", argv[i])==0) {
                                if (strcmp("U", argv[++i])==0)
                                        uplo = MagmaUpper;
                                else
                                        uplo = MagmaLower;
                        }
                        
        }

                if(N < 0)
                        exit(1);
                else if (N > 0) size[0] = size[21] = N;
    }
    else {
        printf("\nUsage: \n");
        printf("  testing_cpotrf_mic -N %d\n\n", 1024);
    }

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
        fprintf( stderr, "magma_get_devices failed: %d\n", err );
        exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
        fprintf( stderr, "magma_queue_create failed: %d\n", err );
        exit(-1);
    }

    /* Allocate memory for the largest matrix */
    N    = size[23];
    lda = ldda = ((N+31)/32) * 32;
        lda = ldda = (ldda % 256 == 0) ? (ldda + 32) : ldda;
    n2   = lda * N;
    TESTING_MALLOC_HOST( hA, magmaFloatComplex, n2 );
    TESTING_MALLOC_HOST( hR, magmaFloatComplex, n2 );
    TESTING_MALLOC_DEV(  dA, magmaFloatComplex, ldda*N );
    
    printf("\n\n");
    printf("  N    CPU GFlop/s (sec)    MIC GFlop/s (sec)    ||R_magma-R_lapack||_F / ||R_lapack||_F\n");
    printf("========================================================================================\n");
    for(i=0; i < 24; i++){
        N   = size[i];
        lda = ldda = ((N+31)/32)*32;
                lda = ldda = (ldda % 256 == 0) ? (ldda + 32) : ldda;
        n2  = lda*N;
        gflops = FLOPS( (float)N ) * 1e-9;
        
        /* Initialize the matrix */
        lapackf77_clarnv( &ione, ISEED, &n2, hA );
        /* Symmetrize and increase the diagonal */
        for( int i = 0; i < N; ++i ) {
            MAGMA_C_SET2REAL( hA(i,i), MAGMA_C_REAL(hA(i,i)) + N );
            for( int j = 0; j < i; ++j ) {
              hA(i, j) = MAGMA_C_CNJG( hA(j,i) );
            }
        }
        lapackf77_clacpy( MagmaFullStr, &N, &N, hA, &lda, hR, &lda );

                /* Warm up to measure the performance */
                magma_csetmatrix( N, N, hA, 0, lda, dA, 0, ldda, queue );
                //magma_cpotrf_mic( MagmaLower, N, dA, 0, ldda, &info, queue );

        /* ====================================================================
           Performs operation using MAGMA 
           =================================================================== */
        magma_csetmatrix( N, N, hA, 0, lda, dA, 0, ldda, queue );
        mic_time = get_time();
                
                if(uplo == MagmaLower)
                        magma_cpotrf_mic( MagmaLower, N, dA, 0, ldda, &info, queue );
                else
                        magma_cpotrf_mic( MagmaUpper, N, dA, 0, ldda, &info, queue );

        mic_time = get_time() - mic_time;
        if (info != 0)
            printf( "magma_cpotrf had error %d.\n", info );

        mic_perf = gflops / mic_time;
        
        /* =====================================================================
           Performs operation using LAPACK 
           =================================================================== */
        cpu_time = get_time();
                if(uplo == MagmaLower)
                        lapackf77_cpotrf( MagmaLowerStr, &N, hA, &lda, &info );
                else
                        lapackf77_cpotrf( MagmaUpperStr, &N, hA, &lda, &info );
        cpu_time = get_time() - cpu_time;
        if (info != 0)
            printf( "lapackf77_cpotrf had error %d.\n", info );
        
        cpu_perf = gflops / cpu_time;
        
        /* =====================================================================
           Check the result compared to LAPACK
           |R_magma - R_lapack| / |R_lapack|
           =================================================================== */
        magma_cgetmatrix( N, N, dA, 0, ldda, hR, 0, lda, queue );
        //matnorm = lapackf77_clange("f", &N, &N, hA, &lda, work);
        matnorm = lapackf77_clanhe("f", "L", &N, hA, &lda, work);
        blasf77_caxpy(&n2, &mz_one, hA, &ione, hR, &ione);
        //diffnorm = lapackf77_clange("f", &N, &N, hR, &lda, work);
        diffnorm = lapackf77_clanhe("f", "L", &N, hR, &lda, work);
        printf( "%5d     %6.2f (%6.2f)     %6.2f (%6.2f)         %e\n", 
                N, cpu_perf, cpu_time, mic_perf, mic_time, diffnorm / matnorm );
       
                fflush(0); 
        if (flag == 1)
            break;
    }

    /* clean up */
    TESTING_FREE_HOST( hA );
    TESTING_FREE_HOST( hR );
    TESTING_FREE_DEV( dA );
    magma_queue_destroy( queue );
    magma_finalize();
}
