/*
 *  -- MAGMA (version 1.2.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     June 2012
 *
 * @generated d Tue Mar 12 12:57:59 2013
 *
 **/
/* includes, system */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* includes, project */
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

#define PRECISION_d
/* Flops formula */
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(n) ( 6. * FMULS_POTRF(n) + 2. * FADDS_POTRF(n) )
#else
#define FLOPS(n) (      FMULS_POTRF(n) +      FADDS_POTRF(n) )
#endif

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing dpotrf_mmic
*/
int main( int argc, char** argv) 
{
	magma_init();

    magma_timestr_t  start, end;
    double      error, flops, mic_perf, cpu_perf;
    double *h_A, *h_R;
    double *d_lA[4];
    magma_int_t N = 0, n2, nb, nk, lda, ldda, n_local, ldn_local;
    //magma_int_t size[10] = {1024,2048,3072,4032,5184,6048,7200,8064,8928,10240};
	magma_int_t size[38] = {960, 1920, 3072, 4032, 4992, 5952, 7104, 8064, 9024, 10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694, 27892, 28776, 29680, 30892, 31784, 32899, 33900, 35170, 36668, 37800, 38896, 40358, 41478, 42680, 43996, 44720};
	
	magma_int_t n_sizes = 37, flag = 0;
    
	magma_int_t i, j, k, info, num_mics0 = 1, num_mics;
	const char *uplo     = MagmaLowerStr;
	double c_neg_one = MAGMA_D_NEG_ONE;
	magma_int_t ione     = 1;
	magma_int_t ISEED[4] = {0, 0, 0, 1};
	double      work[1], matnorm;
	magma_uplo_t magma_uplo;

	magma_queue_t queue;
	magma_device_t device;
	int num = 0;
	magma_err_t err;

	err = magma_get_devices( &device, 1, &num );
	if ( err != MAGMA_SUCCESS or num < 1 ) {
		fprintf( stderr, "magma_get_devices failed: %d\n", err );
		exit(-1);
	}
	err = magma_queue_create( device, &queue );
	if ( err != MAGMA_SUCCESS ) {
		fprintf( stderr, "magma_queue_create failed: %d\n", err );
		exit(-1);
	}

	N = size[n_sizes - 1];
	if (argc != 1){
		for(i = 1; i < argc; i++){        
			if (strcmp("-N", argv[i]) == 0) {
				flag = 1;
				N = atoi(argv[++i]);
				size[0] = size[n_sizes - 1] = N;
			}
			if (strcmp("-NMIC", argv[i]) == 0)
				num_mics0 = atoi(argv[++i]);
			if (strcmp("-UPLO",argv[i]) == 0) {
				if (strcmp("L",argv[++i]) == 0) 
					uplo = MagmaLowerStr;
				else                          
					uplo = MagmaUpperStr;
			}
		}
		if (strcmp(uplo,MagmaLowerStr) == 0)
			printf("\n  testing_dpotrf_mic -N %d -NMIC %d -UPLO L\n\n", (int) N, (int) num_mics0 );
		else
			printf("\n  testing_dpotrf_mic -N %d -NMIC %d -UPLO U\n\n", (int) N, (int) num_mics0 );
	} else {
		printf("\nDefault: \n");
		printf("  testing_dpotrf_mic -N %d:%d -NMIC %d -UPLO L\n\n", (int) size[0], (int) size[n_sizes-1], (int) num_mics0 );
	}

	if( N <= 0 || num_mics0 <= 0 )  {
		printf( " invalid input N = %d Nmic = %d\n", (int) N, (int) num_mics0 );
		exit(1);
	}

	/* Allocate host memory for the matrix */
	n2   = size[n_sizes - 1] * size[n_sizes - 1];
	ldda = ((size[n_sizes - 1] + 31) / 32) * 32;
	ldda = (ldda % 256 == 0) ? ldda + 32 : ldda;
	nb = magma_get_dpotrf_nb(N);

	TESTING_MALLOC_HOST(    h_A, double, n2);
	TESTING_MALLOC_HOST( h_R, double, n2);

	num_mics = num_mics0;
	for(i = 0; i < num_mics; i++){
		n_local = ((N / nb) / num_mics) * nb;
		if (i < (N / nb) % num_mics)
			n_local += nb;
		else if (i == (N / nb) % num_mics)
			n_local += N % nb;

		ldn_local = ((n_local + 31) / 32) * 32;
		ldn_local = (ldn_local % 256 == 0) ? ldn_local + 32 : ldn_local;

		micSetDevice(i + 1);
		TESTING_MALLOC_DEV( d_lA[i], double, ldda * ldn_local );
	}

	printf("  N    CPU GFlop/s    mic GFlop/s    ||R||_F / ||A||_F\n");
	printf("========================================================\n");

	for(i = 0; i < n_sizes; i++){
		N     = size[i];
		lda   = N; 
		n2    = lda * N;
		ldda  = ((N + 31) / 32) * 32;
		ldda = (ldda % 256 == 0) ? ldda + 32 : ldda;
		flops = FLOPS( (double)N ) / 1000000;

		/* Initialize the matrix */
		lapackf77_dlarnv( &ione, ISEED, &n2, h_A );
		/* Symmetrize and increase the diagonal */
		{
			magma_int_t i, j;
			for(i = 0; i < N; i++) {
				MAGMA_D_SET2REAL( h_A[i * lda + i], ( MAGMA_D_REAL(h_A[i * lda + i]) + 1. * N ) );
				for(j = 0; j < i; j++)
					h_A[i * lda + j] = MAGMA_D_CNJG(h_A[j * lda + i]);
			}
		}

		lapackf77_dlacpy( MagmaUpperLowerStr, &N, &N, h_A, &lda, h_R, &lda );

		/* ====================================================================
		   Performs operation using MAGMA 
		   =================================================================== */

		nb = magma_get_dpotrf_nb(N);
		if( num_mics0 > N / nb ) {
			num_mics = N / nb;
			if( N % nb != 0 ) num_mics++;
			printf( " * too many mics for the matrix size, using %d mics\n", (int) num_mics );
		} else {
			num_mics = num_mics0;
		}

		/* distribute matrix to mics */
		if( lapackf77_lsame(uplo, "U") ) {
			for(j = 0; j < N; j += nb){
				k = (j / nb) % num_mics;
				micSetDevice(k + 1);
				nk = min(nb, N - j);
				magma_dsetmatrix( N, nk,
						h_A + j * lda,  0,                     lda,
						d_lA[k] + j / (nb * num_mics) * nb * ldda, 0, ldda, queue);
			}
		} else {
			for(j = 0; j < N; j += nb){
				k = (j / nb) % num_mics;
				n_local = ((N / nb) / num_mics) * nb;
				if (k < (N / nb) % num_mics)
					n_local += nb;
				else if (k == (N / nb) % num_mics)
					n_local += N % nb;
				ldn_local = ((n_local + 31) / 32) * 32;
				ldn_local = (ldn_local % 256 == 0) ? ldn_local + 32 : ldn_local;

				micSetDevice(k + 1);
				nk = min(nb, N - j);
				magma_dsetmatrix( nk, N,
						h_A + j, 0,                     lda,
						d_lA[k] + j / (nb * num_mics) * nb, 0, ldn_local, queue );
			}
		}
		micSetDevice(1);

		if(uplo[0] == 'L')
			magma_uplo = MagmaLower;
		else
			magma_uplo = MagmaUpper;

		/* call magma_dpotrf_mgp */
		start = get_current_time();
		magma_dpotrf_mmic(num_mics, magma_uplo, N, d_lA, ldda, &info, queue);
		end = get_current_time();
		if (info < 0) {
			printf("Argument %d of magma_dpotrf_mmic had an illegal value.\n", (int) -info);
			break;
		} else if (info != 0) {
			printf("magma_dpotrf_mmic returned info = %d\n", (int) info );
			break;
		}
		mic_perf = flops / GetTimerValue(start, end);

		/* gather matrix from mics */
		if( lapackf77_lsame(uplo, "U") ) {
			for(j = 0; j < N; j += nb){
				k = (j / nb) % num_mics;
				micSetDevice(k + 1);
				nk = min(nb, N - j);
				magma_dgetmatrix( N, nk,
						d_lA[k] + j / (nb * num_mics) * nb * ldda, 0, ldda,
						h_R + j * lda, 0,                       lda, queue );
			}
		} else {
			for(j = 0; j < N; j += nb){
				k = (j / nb) % num_mics;
				n_local = ((N / nb) / num_mics) * nb;
				if (k < (N / nb) % num_mics)
					n_local += nb;
				else if (k == (N / nb) % num_mics)
					n_local += N % nb;
				ldn_local = ((n_local + 31) / 32) * 32;
				ldn_local = (ldn_local % 256 == 0) ? ldn_local + 32 : ldn_local;

				micSetDevice(k + 1);
				nk = min(nb, N - j);
				magma_dgetmatrix( nk, N,
						d_lA[k] + j / (nb * num_mics) * nb, 0, ldn_local,
						h_R + j, 0,                      lda, queue );
			}
		}
		micSetDevice(1);

		/* =====================================================================
		   Performs operation using LAPACK 
		   =================================================================== */
		start = get_current_time();
		lapackf77_dpotrf(uplo, &N, h_A, &lda, &info);
		end = get_current_time();
		if (info < 0) {
			printf("Argument %d of dpotrf had an illegal value.\n", (int) -info);
			break;
		} else if (info != 0) {
			printf("lapackf77_dpotrf returned info = %d\n", (int) info );
			break;
		}
		cpu_perf = flops / GetTimerValue(start, end);

		/* =====================================================================
		   Check the result compared to LAPACK
		   =================================================================== */
		matnorm = lapackf77_dlange("f", &N, &N, h_A, &lda, work);
		blasf77_daxpy(&n2, &c_neg_one, h_A, &ione, h_R, &ione);
		error = lapackf77_dlange("f", &N, &N, h_R, &lda, work) / matnorm;

		printf("%5d    %6.2f         %6.2f        %e\n", 
				(int) size[i], cpu_perf, mic_perf, error);
		fflush(0);
		if (flag != 0) break;
	}

	/* Memory clean up */
	TESTING_FREE( h_A );
	TESTING_FREE( h_R );
	for(i = 0; i < num_mics; i++) {
		micSetDevice(i + 1);
		TESTING_FREE_DEV( d_lA[i] );
	}

	/* Shutdown */
	magma_queue_destroy(queue);
	magma_finalize();
}
