/*
    -- MAGMA (version 1.2.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       June 2012

       @generated s Tue Mar 12 12:57:59 2013

*/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

// Flops formula
#define PRECISION_s
#if defined(PRECISION_z) || defined(PRECISION_c)
#define FLOPS(m, n) ( 6.*FMULS_GEQRF(m, n) + 2.*FADDS_GEQRF(m, n) )
#else
#define FLOPS(m, n) (    FMULS_GEQRF(m, n) +    FADDS_GEQRF(m, n) )
#endif

#define Multimics

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing sgeqrf
*/

int main( int argc, char** argv)
{
	magma_init();

	magma_timestr_t       start, end;
	float           error, flops, mic_perf, cpu_perf;
	float           matnorm, work[1];
	float  c_neg_one = MAGMA_S_NEG_ONE;
	float *h_A, *h_R, *tau, *hwork, tmp[1];
	float *d_lA[4];

	/* Matrix size */
	magma_int_t M = 0, N = 0, n2, n_local[4], lda, ldda, lhwork;
	//magma_int_t size[10] = {1024,2048,3072,4032,5184,6016,7040,8064,9088,9984};
	magma_int_t size[38] = {960, 1920, 3072, 4032, 4992, 5952, 7104, 8064, 9024, 10560, 12596, 14560, 16064, 18928, 20160, 21256, 22596, 23560, 24364, 25500, 26694, 27892, 28776, 29680, 30892, 31784, 32899, 33900, 35170, 36668, 37800, 38896, 40358, 41478, 42680, 43996, 44720};

	magma_int_t n_size = 36;

	magma_int_t i, j, d, k, nk, info, min_mn, flag;
	int max_num_mics, num_mics = 1;

	magma_int_t ione     = 1;
	magma_int_t ISEED[4] = {0,0,0,1};

	magma_queue_t queue;
	magma_device_t device;
	int num = 0;
	magma_err_t err;

	err = magma_get_devices( &device, 1, &num );
	if ( err != MAGMA_SUCCESS or num < 1 ) {
		fprintf( stderr, "magma_get_devices failed: %d\n", err );
		exit(-1);
	}
	err = magma_queue_create( device, &queue );
	if ( err != MAGMA_SUCCESS ) {
		fprintf( stderr, "magma_queue_create failed: %d\n", err );
		exit(-1);
	}

	flag = 0;
	if (argc != 1){
		for(i = 1; i < argc; i++){
			if (strcmp("-N", argv[i]) == 0) {
				N = atoi(argv[++i]);
				flag = 1;
			}
			else if (strcmp("-M", argv[i]) == 0) {
				M = atoi(argv[++i]);
				flag = 1;
			}
			else if (strcmp("-NMIC", argv[i]) == 0)
				num_mics = atoi(argv[++i]);
		}
	}

	if(flag != 0) {
		if (M > 0 && N > 0 && num_mics > 0)
			printf("  testing_sgeqrf_mic -M %d -N %d -Nmic %d\n\n", (int) M, (int) N, (int) num_mics);
		else
		{
			printf("\nUsage: \n");
			printf("  testing_sgeqrf_mic -M %d -N %d -Nmic %d\n\n", 
					1024, 1024, 1);
			exit(1);
		}
	}
	else {
		printf("\nUsage: \n");
		printf("  testing_sgeqrf_mic -M %d -N %d -Nmic %d\n\n", 1024, 1024, 1);
		M = N = size[n_size - 1];
	}

	ldda   = ((M + 31) / 32) * 32;
	ldda = (ldda % 256 == 0) ? ldda + 32 : ldda;

	n2     = M * N;
	min_mn = min(M, N);

	magma_int_t nb  = magma_get_sgeqrf_nb(M);
	//cudaGetDeviceCount(&max_num_mics);
	//if (num_mics > max_num_mics){
	//  printf("More mics requested than available. Have to change it.\n");
	//  num_mics = max_num_mics;
	//}
	printf("Number of mics to be used = %d\n", (int) num_mics);

	/* Allocate host memory for the matrix */
	TESTING_MALLOC_HOST(    tau, float, min_mn );
	TESTING_MALLOC_HOST(    h_A, float, n2     );
	TESTING_MALLOC_HOST(    h_R, float, n2     );

	for(i = 0; i < num_mics; i++){      
		n_local[i] = ((N / nb) / num_mics) * nb;
		if (i < (N / nb) % num_mics)
			n_local[i] += nb;
		else if (i == (N / nb) % num_mics)
			n_local[i] += N % nb;

		micSetDevice(i + 1);
		if(n_local[i] > 0)
			TESTING_MALLOC_DEV(  d_lA[i], float, ldda * n_local[i] );
		printf("device %2d n_local = %4d\n", (int) i, (int) n_local[i]);  
	}

	micSetDevice(1);

	lhwork = -1;
	lapackf77_sgeqrf(&M, &N, h_A, &M, tau, tmp, &lhwork, &info);
	lhwork = (magma_int_t)MAGMA_S_REAL( tmp[0] );

	TESTING_MALLOC_HOST( hwork, float, lhwork );

	printf("  M     N   CPU GFlop/s   mic GFlop/s    ||R||_F / ||A||_F\n");
	printf("==========================================================\n");
	
	for(i = 0; i < n_size; i++){
		if (flag == 0){
			M = N = size[i];
		}
		min_mn = min(M, N);
		lda   = M;
		n2    = lda * N;
		ldda  = ((M + 31) / 32) * 32;
		ldda = (ldda % 256 == 0) ? ldda + 32 : ldda;
		flops = FLOPS( (float)M, (float)N ) / 1000000;

		/* Initialize the matrix */
		lapackf77_slarnv( &ione, ISEED, &n2, h_A );
		lapackf77_slacpy( MagmaUpperLowerStr, &M, &N, h_A, &lda, h_R, &lda );

		/* =====================================================================
		   Performs operation using LAPACK
		   =================================================================== */
		start = get_current_time();
		lapackf77_sgeqrf(&M, &N, h_A, &M, tau, hwork, &lhwork, &info);
		end = get_current_time();
		if (info < 0)
			printf("Argument %d of lapack_sgeqrf had an illegal value.\n", (int) -info);

		cpu_perf = flops / GetTimerValue(start, end);

		/* ====================================================================
		   Performs operation using MAGMA
		   =================================================================== */
		//magmablas_ssetmatrix_1D_bcyclic(M, N, h_R, lda, d_lA, ldda, num_mics, nb);
		for(j = 0; j < N; j += nb ) {
			d = (j / nb) % num_mics;
			micSetDevice(d + 1);
			nk = min(nb, N - j);
			magma_ssetmatrix( M, nk,
					h_R + j * lda, 0, lda,
					d_lA[d] + j / (nb * num_mics) * nb * ldda, 0, ldda, queue);
		}

		micSetDevice(1);

		start = get_current_time();
		magma_sgeqrf2_mmic( num_mics, M, N, d_lA, ldda, tau, &info, queue);
		end = get_current_time();

		if (info < 0)
			printf("Argument %d of magma_sgeqrf2 had an illegal value.\n", (int) -info);

		mic_perf = flops / GetTimerValue(start, end);

		/* =====================================================================
		   Check the result compared to LAPACK
		   =================================================================== */
		
		//magmablas_sgetmatrix_1D_bcyclic(M, N, d_lA, ldda, h_R, lda, num_mics, nb);
		for( j = 0; j < N; j += nb ) {
			d = (j / nb) % num_mics;
			micSetDevice( d + 1 );
			nk = min(nb, N - j);
			magma_sgetmatrix( M, nk,
					d_lA[d] + j / (nb * num_mics) * nb * ldda, 0, ldda,
					h_R + j * lda, 0, lda, queue);
		}

		micSetDevice(1);

		matnorm = lapackf77_slange("f", &M, &N, h_A, &M, work);
		blasf77_saxpy(&n2, &c_neg_one, h_A, &ione, h_R, &ione);
		error = lapackf77_slange("f", &M, &N, h_R, &M, work) / matnorm;

		printf("%5d %5d  %6.2f         %6.2f        %e\n",
				(int) M, (int) N, cpu_perf, mic_perf, error);

		fflush(0);

		if (flag != 0)
			break;
	}

	/* Memory clean up */
	TESTING_FREE( tau );
	TESTING_FREE( h_A );
	TESTING_FREE( hwork );
	TESTING_FREE( h_R );

	for(i = 0; i < num_mics; i++){
		micSetDevice(i + 1);
		if(d_lA[i] != NULL)
			TESTING_FREE_DEV( d_lA[i] );
	}

	/* Shutdown */
	magma_queue_destroy(queue);
	magma_finalize();
}
